import os
import re
import sys
import time
import signal
import logging
import argparse
import functools
import subprocess
import tracemalloc
from typing import List, Union, Dict, Callable, Tuple, Any
from pathlib import Path

import pika
import pendulum
from pycapo import CapoConfig

from .wf_event import WorkflowEvent, EventType
from ._version import ___version___ as VERSION

WORKFLOW_STATUS_EXCH = 'workspaces.workflow-service.workflow-status'

logger = logging.getLogger('wf_monitor')
logger.setLevel(logging.INFO)
logger.addHandler(logging.StreamHandler(sys.stdout))


def log_decorator_factory(msg: str):
    """
    Factory that produces log decorators with a given message
    :param msg: Message to be logged
    :return: Log decorator
    """
    def log_decorator(f: Callable) -> Callable:
        """
        Decorator that logs a message before and after a function is called
        :param f: Function to be decorated
        :return: Wrapper function
        """
        @functools.wraps(f)
        def wrapper(*args: Tuple, **kwargs: Dict) -> Any:
            """
            Decorator boilerplate: wrapper function for log decorator
            :param args: args for function f
            :param kwargs: kwargs for function f
            :return: Return value of the function
            """
            logger.info(msg)
            result = f(*args, **kwargs)
            logger.info("Done.")
            return result
        return wrapper
    return log_decorator


def timeout_handler(signum: int, frame: tracemalloc.Frame) -> None:
    """
    Signal handler function for handling a SIGALRM that arrives when WorkflowMonitor.read_log()
    times out.
    :param signum: Signal number (would be signal.SIGALRM in this case)
    :param frame: Stack frame the signal was sent from
    """
    msg = 'Timeout elapsed when attempting to read HTCondor log file. The job either did ' \
               'not complete or the path to the log was not entered correctly.'
    raise TimeoutError(msg)


# Timeout handling solution found here:
# https://stackoverflow.com/questions/492519/timeout-on-a-function-call
signal.signal(signal.SIGALRM, timeout_handler)


def get_job_name(body: str) -> str:
    """
    Parses a log entry body string for the name of the job currently executing
    :param body: Body of the log entry
    :return: Job name
    """
    r_body_parse = r'[a-zA-Z0-9-_ ]+: (?P<jobname>[a-zA-Z0-9-_]+)'
    body_match = re.search(r_body_parse, body)
    return body_match.group('jobname')


def get_year(logfile_path: Path) -> str:
    """
    Parses year from log file creation date
    :param logfile_path: Path to log file
    :return: Year log was created as a string
    """
    status = os.stat(logfile_path)
    date = pendulum.from_timestamp(status.st_ctime)
    return str(date.year)


def fmt_timestamp(timestamp: str, year: str) -> str:
    """
    Format timestamp string into ISO 8601 format
    2020-08-27T21:45:09+07:00
    :param timestamp: Unformatted timestamp string
    :param year: Year HTCondor log was created as a string
    :return: Formatted timestamp string
    """
    r_nums = r'[0-9]{2}'
    r_timestamp = rf'(?P<mon>{r_nums})/(?P<d>{r_nums}) ' \
                       rf'(?P<h>{r_nums}):(?P<m>{r_nums}):(?P<s>{r_nums})'
    match = re.search(r_timestamp, timestamp)

    return '{year}-{mon}-{day}T{h}:{m}:{s}+07:00'.format(
        year=year,
        mon=match.group('mon'),
        day=match.group('d'),
        h=match.group('h'),
        m=match.group('m'),
        s=match.group('s')
    )


def get_retval(body: str) -> Union[int, None]:
    """
    Parse return value from job terminated event
    :param body: Body of event log entry
    :return: Return value as an int
    """
    r_retval = r'return value (?P<retval>[0-9]+)'

    retval = re.search(r_retval, body)
    if retval:
        return int(retval.group('retval'))
    else:
        return


class WorkflowMonitor:
    def __init__(self, logfile_path: str, timeout: int = 60):
        self.logfile_path = Path(logfile_path)

        try:
            self.log = self.read_htcondor_log(timeout)
        except TimeoutError as e:
            # read_htcondor_log() timed out
            logger.error(e)
            quit(-1)

        self.events = self.parse_log()

    @log_decorator_factory('Reading HTCondor log...')
    def read_htcondor_log(self, timeout: int) -> str:
        """
        Waits for HTCondor job to terminate and reads the resulting log file
        :param logfile_path: Path to log file
        :param timeout: Time in seconds to wait for the job to finish
        :return: String contents of the log file
        """
        signal.alarm(timeout)

        while not os.path.exists(self.logfile_path):
            logger.info('Attempting to read log file...')
            time.sleep(3)

        logger.info('Log file found.')
        with open(self.logfile_path, 'r') as f:
            return f.read()

    @log_decorator_factory('Parsing HTCondor log...')
    def parse_log(self) -> List[WorkflowEvent]:
        """
        Parse log for relevant details:
        -
        :return: List of WorkflowEvent objects
        """
        # Regex that matches an event entry in the HTCondor log file.
        r_htc_event = r'(?P<eventnum>[0-9]{3}) \((?P<jobnum>[0-9]{4})\.[0-9]{3}\.[0-9]{3}\) ' \
                           r'(?P<timestamp>[0-9]{2}/[0-9]{2} [0-9]{2}:[0-9]{2}:[0-9]{2}) ' \
                           r'(?P<desc>[a-zA-Z0-9 ?&:=<>\-._]+)\n' \
                           r'(?P<body>[\t ]+[^\.]*\n)*' \
                           r'...'
        jobs = {}
        events = []
        if sum(1 for _ in re.finditer(r_htc_event, self.log)) == 0:
            raise ValueError('HTCondor log file not well-formatted. Was the file edited?')
        for match in re.finditer(r_htc_event, self.log):
            job_id = match.group('jobnum')
            if match.group('eventnum') == '000':
                # Populate jobs dict with job name
                jobs[job_id] = get_job_name(match.group('body'))

            try:
                event_type = EventType(int(match.group('eventnum')))
            except ValueError:
                event_type = EventType.OTHER

            if event_type == EventType.TERMINATED:
                retval = get_retval(match.group('body'))
            else:
                retval = None

            year = get_year(self.logfile_path)
            events.append(WorkflowEvent(
                job_name=jobs[job_id],
                job_id=job_id,
                event_type=event_type,
                timestamp=fmt_timestamp(match.group('timestamp'), year),
                log=match.group(0),
                retval=retval
            ))

        return events

    def __str__(self):
        return f'WorkflowMonitor, monitoring {self.logfile_path} that has events {self.events}'

    def __repr__(self):
        return f'{self.__class__.__name__} ({self.logfile_path}, {self.events})'


_DESCRIPTION = 'Workspaces workflow monitor, version {}. Monitor execution of a workflow from ' \
               'the command line.'


def make_arg_parser() -> argparse.ArgumentParser:
    """
    Initialize argument parser for command-line arguments
    :return: Argument parser
    """
    parser = argparse.ArgumentParser(
        description=_DESCRIPTION.format(VERSION),
        formatter_class=argparse.RawTextHelpFormatter
    )
    parser.add_argument('-P', '--profile', action='store', required=False, default='nmtest',
                        help='profile name to use; e.g. test, production; DEFAULT: nmtest')
    parser.add_argument('log_path', action='store', help='path to the workflow\'s HTCondor log')

    return parser


@log_decorator_factory('Establishing connection to the AMQP server...')
def make_amqp_connection(profile: str) -> pika.BlockingConnection:
    """
    Initialize connection to AMQP server
    :param profile: Capo profile to use
    :return: Established connection
    """
    capo_cfg = CapoConfig(profile=profile)

    hostname = capo_cfg.getstring('edu.nrao.archive.configuration.AmqpServer.hostname')
    username = capo_cfg.getstring('edu.nrao.archive.configuration.AmqpServer.username')
    password = capo_cfg.getstring('edu.nrao.archive.configuration.AmqpServer.password')

    return pika.BlockingConnection(
        pika.ConnectionParameters(
            hostname,
            credentials=pika.PlainCredentials(
                username,
                password
            )
        )
    )


@log_decorator_factory('Sending event data to the channel...')
def send_event_data(connection: pika.BlockingConnection, event: WorkflowEvent):
    """
    Takes in a JSON-formatted string and sends it off to the workflow status exchange
    :param event: JSON string of event metadata
    """
    # Send data to exchange
    channel = connection.channel()
    channel.exchange_declare(WORKFLOW_STATUS_EXCH, 'topic', durable=True)
    return channel.basic_publish(
        exchange=WORKFLOW_STATUS_EXCH,
        routing_key=f'{event.job_name}.{event.job_id}.{event.type.name.lower()}',
        body=event.json()
    )


def main():
    # Parse command-line args
    args = make_arg_parser().parse_args()
    connection = make_amqp_connection(args.profile)
    monitor = WorkflowMonitor(args.log_path)

    for event in monitor.events:
        send_event_data(connection, event)
